@@ -2,9 +2,9 @@
 import re
 import logging
 import socket
-from ftplib import FTP, error_temp
 from pathlib import Path
 from functools import partial
+from ftplib import FTP, error_temp, error_perm
 
 from tqdm import tqdm
 
@@ -46,9 +46,13 @@ class FTPParser:
         The url for the FTP connection.
     max_depth : int, optional
         The maximum resursion depth when looking for files.
+    max_reconnects : int, optional
+        The maximum number of reconnects to attempt during downloads.
+    timeout : float, optional
+        The maximum amount of time to wait for a response from the server.
     """
 
-    def __init__(self, url, max_depth=4, max_reconnects=10):
+    def __init__(self, url, max_depth=4, max_reconnects=10, timeout=10.0):
         """Initialize an FTPParser"""
         if not url.startswith("ftp://"):
             raise ValueError("The URL does not appear to be an FTP server")
@@ -57,31 +61,59 @@ class FTPParser:
         self.connection = None
         self.max_depth = max_depth
         self.max_reconnects = max_reconnects
+        self.timeout = timeout
         self._files = None
         self._dirs = None
         self._depth = 0
 
-        # The directory at the time of disconnect:
-        self.last_pwd = None
-
-    def connect(self):
+    def _connect(self):
         """Connect to the FTP server"""
         if self.connection is not None and self.connection.file is None:
             self.connection.close()
             self.connection = None
 
         if self.connection is None:
-            self.connection = FTP()
+            self.connection = FTP(timeout=self.timeout)
             self.connection.connect(self.server)
             self.connection.login()
             self.connection.cwd(self.path)
 
+    def connect(self):
+        """Connect to the FTP server, with reconnects on failure."""
+        self._with_reconnects(self._connect)
+
     def quit(self):
         """Close the connection."""
         if self.connection is not None:
-            self.connection.quit()
+            self.connection.close()
             self.connection = None
 
+    def _with_reconnects(self, func, *args, **kwargs):
+        """Try and execute a function, reconnecting on failure."""
+        for _ in range(self.max_reconnects):
+            try:
+                self._connect()
+                return func(*args, **kwargs)
+
+            except (
+                ConnectionRefusedError,
+                ConnectionResetError,
+                socket.timeout,
+                socket.gaierror,
+                socket.herror,
+                error_temp,
+                error_perm,
+                EOFError,
+                OSError,
+            ) as err:
+                self.quit()
+                last_err = err
+
+        raise error_temp(
+            f"Failed after {self.max_reconnects} reconnect(s), "
+            f"the last error was: {last_err}"
+        )
+
     def _download_file(self, remote_file, out_file, force_, silent):
         """Download a single file.
 
@@ -99,9 +131,7 @@ class FTPParser:
         force_ : bool
             Force the file to be redownloaded, even if it exists.
         """
-        if self.connection is None:
-            self.connect()
-
+        self.connect()
         size = self.connection.size(remote_file)
         pbar = tqdm(
             desc=str(remote_file),
@@ -113,7 +143,7 @@ class FTPParser:
             leave=False,
             disable=silent,
         )
-        last_err = None
+
         mode = "wb+" if force_ else "ab+"
         with out_file.open(mode) as out:
             start_pos = out.tell()
@@ -125,42 +155,35 @@ class FTPParser:
                 return
 
             # Download file if not:
-            write = partial(write_file, fhandle=out, pbar=pbar)
-            for _ in range(0, self.max_reconnects):
-                try:
-                    if self.connection is None:
-                        self.connect()
+            self._with_reconnects(
+                self._transfer_file,
+                fname=remote_file,
+                fhandle=out,
+                pbar=pbar,
+            )
 
-                    self.connection.retrbinary(
-                        f"RETR {remote_file}",
-                        write,
-                        rest=out.tell(),
-                    )
-                    pbar.close()
-                    return
+        self.quit()
 
-                except (
-                    ConnectionRefusedError,
-                    socket.timeout,
-                    socket.gaierror,
-                    socket.herror,
-                    error_temp,
-                    EOFError,
-                    OSError,
-                ) as err:
-                    self.connection.close()
-                    self.connection = None
-                    last_err = err
+    def _transfer_file(self, fname, fhandle, pbar):
+        """Perform the actual file transfer.
 
-        raise error_temp(
-            f"Failed after {self.max_reconnects} reconnect(s), "
-            f"the last error was: {last_err}"
-        )
+        Parameters
+        ----------
+        fname : str
+            The remote file name.
+        fhandle : file object
+            The opened file object where the data will be written.
+        pbar : tqdm.tqdm
+            The tqdm progress bar to update.
+        """
+        write = partial(write_file, fhandle=fhandle, pbar=pbar)
+        self.connection.retrbinary(f"RETR {fname}", write, rest=fhandle.tell())
+        pbar.close()
 
     def _get_files(self):
         """Recursively list files from the FTP connection."""
         self.connect()
-        self._files, self._dirs = self._parse_files()
+        self._files, self._dirs = self._with_reconnects(self._parse_files)
         self._depth = 0
         self.quit()
 
@@ -219,6 +242,7 @@ class FTPParser:
         """List the files form the FTP connection"""
         if self._files is None:
             self._get_files()
+            self._files.sort()
 
         return self._files
 
@@ -227,6 +251,7 @@ class FTPParser:
         """List the directories form the FTP connection"""
         if self._dirs is None:
             self._get_files()
+            self._dirs.sort()
 
         return self._dirs
 
